@@ -5,10 +5,13 @@ from queue import Queue
 from threading import Thread
 from typing import Sequence
 from functools import lru_cache
+import os
+import subprocess
+from time import time, sleep
 
 from .csobject import CsObject
-from .protocol import rpc
-from .util import BrokenPromiseException
+from .protocol import rpc, MAGIC_NUMBER, UNITON_VERSION
+from .csutil import BrokenPromiseException
 
 TLIST = (
   (int, rpc.INT32, struct.Struct("i")),
@@ -35,8 +38,14 @@ def recvall(sock, n):
   return data
 
 
-import subprocess
-from time import time
+
+def get_free_port():
+  """Not 100% guaranteed to be free but good enough"""
+  s = socket.socket()
+  s.bind(('', 0))
+  port = s.getsockname()[1]
+  s.close()
+  return port
 
 
 class UnityProc:
@@ -54,13 +63,16 @@ class UnityProc:
     atexit.register(self.close)
 
     if path is not None:
-      # TODO: find free port and set as env variable
+      host = '127.0.0.1'
+      port = get_free_port() if port is None else port
+      env = os.environ.copy()
+
+      env["UNITONPORT"] = str(port)
       # TODO: allow to set cmd line args
       # TODO: figure out how to set -screen-height -screen-width robustly
-      self._proc = subprocess.Popen([path])
+      self._proc = subprocess.Popen([path], env=env)
       # self._proc = subprocess.Popen([path, '-batchmode'])  # while this still renders on Mac OS it's much slower
-      host = '127.0.0.1'
-      port = 11001 if port is None else port
+
     else:
       host = '127.0.0.1' if host is None else host
       port = 11000 if port is None else port
@@ -82,20 +94,30 @@ class UnityProc:
           raise RuntimeError(f"Process {path} died before Uniton could connect!")
         elif time() > timeout:
           raise ConnectionRefusedError(f"Is {path} missing Uniton?")
+        else:
+          sleep(0.1) # wait, then repeat
 
     self._sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
 
     # version compatibility check
-    remote_version, = struct.unpack('i', recvall(self._sock, 4))
+    magic_number, = struct.unpack('i', recvall(self._sock, 4))
+    if magic_number != MAGIC_NUMBER:
+      raise ConnectionError("Could not connect.")
 
-    if remote_version not in rpc.compatible_versions:
-      # TODO: do better versioning
-      raise AttributeError(f"Remote version {remote_version} is not a compatible version. Must be in {rpc.compatible_versions}")
-      import urllib.request, json
-      with urllib.request.urlopen("https://pypi.org/pypi/uniton/json") as url:
-        data = json.loads(url.read().decode())
-        all_versions = data["info"]["releases"].keys()
-        # TODO: work in progress
+
+    remote_version_tuple = struct.unpack('iii', recvall(self._sock, 4 * 3))
+    self._remote_version = ".".join(str(x) for x in remote_version_tuple)
+    local_version_macro = ".".join(UNITON_VERSION.split(".")[:2])
+    remote_version_macro = ".".join(self._remote_version.split(".")[:2])
+    if remote_version_macro != local_version_macro:
+      # try:
+      #   import urllib.request, json
+      #   with urllib.request.urlopen("https://pypi.org/pypi/uniton/json") as url:
+      #     data = json.loads(url.read().decode())
+      #     all_versions = data["info"]["releases"].keys()
+      #     best_compatible_local_version = sorted(v for v in all_versions if v.startswith(".".join(remote_version_tuple[:2])))[-1]
+      # finally:
+      raise ConnectionError(f'Remote Uniton version {self._remote_version} is not compatible with local version {UNITON_VERSION}. Change remote version to {local_version_macro}.* or local version to {remote_version_macro}.* via\n\npip install "uniton=={remote_version_macro}.*" --force-reinstall\n')
 
 
     # init object management
@@ -176,7 +198,7 @@ class UnityProc:
 
         # get data
         d = recvall(self._sock, n)
-        if d is None: break
+        if d is None: break  # TODO: shouldn't this be an error?
         exc, = struct.unpack_from('i', d, 0)
         if exc:
           # raise RemoteException(self.deserialize(d[4:]))
